package pku.edu.hocra.graph.ds;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class Graph implements Serializable {

	private class Edge implements Serializable {
		int src, tgt, edgeType;
		Edge(int src, int tgt, int edgeType) {
			this.src = src;
			this.tgt = tgt;
			this.edgeType = edgeType;
		}
	}
	
	private ArrayList<ArrayList<Edge>> forwardEdges;
	private ArrayList<ArrayList<Edge>> backwardEdges;
	private ArrayList<Integer> nodeTypeArray;
	//private ArrayList<Boolean> isLibraryCode;
	private ArrayList<String> node2Method;
	private ArrayList<String> node2Unit;
	private ArrayList<Integer> paraId;
	
	private int n = 0;
	private int m = 0;
	
	private Iterator<Edge> edgeIterator;
	private Edge currentEdge;
	
	private int maxParenthesis = 0;
	
	public Graph() {
		forwardEdges = new ArrayList<ArrayList<Edge>>();
		backwardEdges = new ArrayList<ArrayList<Edge>>();
		nodeTypeArray = new ArrayList<Integer>();
		//isLibraryCode = new ArrayList<Boolean>();
		node2Method = new ArrayList<String>();
		node2Unit = new ArrayList<String>();
		paraId = new ArrayList<Integer>();
	}
	
	public int getN() {
		return n;
	}
	
	public int getNumberOfEdges() {
		return m;
	}
	
	public void addNode(int nodeType, String methodStr, String unitStr, int para) {
		forwardEdges.add(new ArrayList<Edge>());
		nodeTypeArray.add(nodeType);
		//isLibraryCode.add(libraryCode);
		node2Method.add(methodStr);
		node2Unit.add(unitStr);
		paraId.add(para);
		n ++;
	}
	
	public void addEdge(int src, int tgt, int edgeType) {
		edgeType = IPathType_1.convert2PathType(edgeType);
		if (edgeType > maxParenthesis) maxParenthesis = edgeType;
		else if (-edgeType > maxParenthesis) maxParenthesis = -edgeType;
		Edge e = new Edge(src, tgt, edgeType);
		forwardEdges.get(src).add(e);
		//backwardEdges.get(tgt).add(e);
		m ++;
	}
	
	public void initializeForwardEdgeIterator(int src) {
		edgeIterator = forwardEdges.get(src).iterator();
	}
	
	public void initializeBackwardEdgeIterator(int src) {
		edgeIterator = backwardEdges.get(src).iterator();
	}
	
	public boolean hasNextEdge() {
		if (edgeIterator.hasNext()) {
			currentEdge = edgeIterator.next();
			return true;
		}
		else
			return false;
	}
	
	public int getForwardEdgeTarget() {
		return currentEdge.tgt; 
	}
	
	public int getBackwardEdgeSource() {
		return currentEdge.src;
	}
	
	public int getEdgeType() {
		return currentEdge.edgeType; 
	}
	
	public int getNodeType(int node) {
		return nodeTypeArray.get(node);
	}
	
	public String getMethodStr(int node) {
		return node2Method.get(node);
	}
	
	public String getUnitStr(int node) {
		return node2Unit.get(node);
	}
	
	public int getParaId(int node) {
		return paraId.get(node);
	}
	
	public int getMaxParenthesis() {
		return maxParenthesis;
	}
	
//	public boolean getNodeIsLibraryCode(int node) {
//		return isLibraryCode.get(node);
//	}
	
	public void printStatistics() {
		System.out.println("n:" + n + ";m:" + m);
	}
	
	public void constructBackwardEdges() {
		backwardEdges = new ArrayList<ArrayList<Edge>>();
		for (int i = 0; i < n; i++)
			backwardEdges.add(new ArrayList<Edge>());
		for (ArrayList<Edge> ae : forwardEdges) {
			for (Edge e : ae) {
				backwardEdges.get(e.tgt).add(e);
			}
		}
		
	}

}
