//
// Created by Wayne on 16/1/9.
//

#include "application_graph.h"

using std::queue;
using std::ifstream;
using std::stringstream;

const vector<const conditional_edge*>* application_graph::get_conds_cbcla(library_graph *lib_gr) const {
    int lib_nodes = lib_gr->get_num_nodes();
    unordered_map<int, vector<const conditional_edge*>> rec1, rec2;
    unordered_set<int> inputs, outputs;
    /*for (auto e : this->edges) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
    }*/
    for (auto e : this->parens) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
        if (src < lib_nodes && tgt < lib_nodes) {
            int sym = e->get_symbol();
            if (sym < 0) sym = -sym;
            size_t tmp = 1UL << 62 | sym;
            if (this->implemeted.count(tmp) != 0) {
                outputs.insert(src);
                inputs.insert(tgt);
            }
        }
    }
    for (int src : inputs) {
        for (auto ce : lib_gr->edges_from(src, GRAM_R)) {
            rec1[ce->get_target()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_from(src, GRAM_R0)) {
            rec1[ce->get_target()].push_back(ce);
        }
    }
    for (int tgt : outputs) {
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L)) {
            rec2[ce->get_source()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L0)) {
            rec2[ce->get_source()].push_back(ce);
        }
    }
    for (int n = 0; n < lib_nodes; n++) {
        if (rec1.count(n) != 0 && rec2.count(n) != 0) {
            auto &ls = rec1.at(n);
            auto &rs = rec2.at(n);
            for (auto ce1 : ls) {
                int src = ce1->get_source();
                auto premise = ce1->get_premise();
                for (auto ce2 : rs) {
                    int tgt = ce2->get_target();
                    auto new_premise = premise->merge(ce2->get_premise());
                    if (!lib_gr->is_redundant(src, tgt, GRAM_D, new_premise)) {
                        auto new_edge = new conditional_edge(src, tgt, GRAM_D, new_premise);
                        lib_gr->add_record(new_edge);
                        lib_gr->add_edge(new_edge);
                    } else {
                        delete new_premise;
                    }
                }
            }
        }
    }

    auto start = clock();

    vector<const conditional_edge*> *res = new vector<const conditional_edge*>();
    unordered_set<const conditional_edge*> hash;

    for (int src : inputs) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : lib_gr->edges_from(src, s)) {
                if (outputs.count(ce->get_target()) != 0) {
                    if (hash.count(ce) != 0) continue;
                    hash.insert(ce);
                    res->push_back(ce);
                    auto premise = ce->get_premise();
                    int sz = premise->size();
                    bool chk = true;
                    for (int i = 0; i < sz && chk; i++) {
                        auto fr = premise->get_one(i);
                        if ((fr >> 62) & 1) {
                            chk = (this->implemeted.count(fr) != 0);
                        } else {
                            edge e = edge::decode(fr);
                            chk = inputs.count(e.get_source()) && outputs.count(e.get_target());
                        }
                    }
                    if (chk) {
                        res->push_back(ce);
                    }
                }
            }
        }
    }

    auto end = clock();
    cerr << "Select time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    return res;
}

const vector<const conditional_edge*>* application_graph::get_conds_f(library_graph *lib_gr) const {
    int lib_nodes = lib_gr->get_num_nodes();
    unordered_map<int, vector<const conditional_edge*>> rec1, rec2;
    unordered_set<int> inputs, outputs;
    /*for (auto e : this->edges) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
    }*/
    for (auto e : this->parens) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
    }
    for (int src : inputs) {
        for (auto ce : lib_gr->edges_from(src, GRAM_R)) {
            rec1[ce->get_target()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_from(src, GRAM_R0)) {
            rec1[ce->get_target()].push_back(ce);
        }
    }
    for (int tgt : outputs) {
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L)) {
            rec2[ce->get_source()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L0)) {
            rec2[ce->get_source()].push_back(ce);
        }
    }
    for (int n = 0; n < lib_nodes; n++) {
        if (rec1.count(n) != 0 && rec2.count(n) != 0) {
            auto &ls = rec1.at(n);
            auto &rs = rec2.at(n);
            for (auto ce1 : ls) {
                int src = ce1->get_source();
                auto premise = ce1->get_premise();
                for (auto ce2 : rs) {
                    int tgt = ce2->get_target();
                    auto new_premise = premise->merge(ce2->get_premise());
                    if (!lib_gr->is_redundant(src, tgt, GRAM_D, new_premise)) {
                        auto new_edge = new conditional_edge(src, tgt, GRAM_D, new_premise);
                        lib_gr->add_record(new_edge);
                        lib_gr->add_edge(new_edge);
                    } else {
                        delete new_premise;
                    }
                }
            }
        }
    }

    auto start = clock();

    vector<const conditional_edge*> *res = new vector<const conditional_edge*>();
    unordered_set<const conditional_edge*> hash;

    for (int src : inputs) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : lib_gr->edges_from(src, s)) {
                if (outputs.count(ce->get_target()) != 0) {
                    if (hash.count(ce) != 0) continue;
                    hash.insert(ce);
                    res->push_back(ce);
                    auto premise = ce->get_premise();
                    int sz = premise->size();
                    bool chk = true;
                    for (int i = 0; i < sz && chk; i++) {
                        auto fr = premise->get_one(i);
                        if ((fr >> 62) & 1) {
                            chk = (this->implemeted.count(fr) != 0);
                        } else {
                            edge e = edge::decode(fr);
                            chk = inputs.count(e.get_source()) && outputs.count(e.get_target());
                        }
                    }
                    if (chk) {
                        res->push_back(ce);
                    }
                }
            }
        }
    }

    auto end = clock();
    cerr << "Select time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    return res;
}

/*static premise_base* get_premise(const vector<size_t> &pr) {
    int sz = pr.size();
    if (sz == 0) {
        return new premise_0;
    } else if (sz == 1) {
        return new premise_1(pr[0]);
    } else if (sz == 2) {
        return new premise_2(pr[0], pr[1]);
    } else if (sz == 3) {
        return new premise_3(pr[0], pr[1], pr[2]);
    } else if (sz == 4) {
        return new premise_4(pr[0], pr[1], pr[2], pr[3]);
    } else if (sz == 5) {
        return new premise_5(pr[0], pr[1], pr[2], pr[3], pr[4]);
    } else {
        return new premise_f(pr);
    }
}

static const conditional_edge* dfs1(const conditional_edge *ce, const unordered_set<size_t> &implemented, hocra_mapping *mapping);
static bool dfs2(const edge *e, const unordered_set<size_t> &implemented, hocra_mapping *mapping);

static const conditional_edge* dfs1(const conditional_edge *ce, const unordered_set<size_t> &implemented, hocra_mapping *mapping) {
    if (ce->get_premise()->size() == 0) {
        return ce;
    } else {
        auto premise = ce->get_premise();
        vector<size_t> new_premise;
        int sz = premise->size();
        for (int i = 0; i < sz; i++) {
            auto fr = premise->get_one(i);
            if ((fr >> 62) & 1) {
                if (implemented.count(fr) == 0) {
                    return ce;
                }
            } else {
                edge e = edge::decode(fr);
                if (!dfs2(&e, implemented, mapping)) {
                    new_premise.push_back(fr);
                }
            }
        }
        if (sz == new_premise.size()) {
            return ce;
        } else {
            return new conditional_edge(ce->get_source(), ce->get_target(), ce->get_symbol(), get_premise(new_premise));
        }
    }
}

static bool dfs2(const edge *e, const unordered_set<size_t> &implemented, hocra_mapping *mapping) {
    assert(!is_paren(e->get_symbol()));
    if (!mapping->has(e->get_source(), e->get_target(), e->get_symbol())) {
        return false;
    } else {
        for (auto &tuple : mapping->get(e->get_source(), e->get_target(), e->get_symbol())) {
            int sz = 0;
            for (auto ce : tuple) {
                sz += dfs1(ce, implemented, mapping)->get_premise()->size();
            }
            if (sz == 0) {
                return true;
            }
        }
        return false;
    }
}*/

const vector<const conditional_edge*>* application_graph::get_conds_k(int k, library_graph *lib_gr, hocra_mapping *mapping) const {
    int lib_nodes = lib_gr->get_num_nodes();
    unordered_map<int, vector<const conditional_edge*>> rec1, rec2;
    unordered_set<int> inputs, outputs;
    /*for (auto e : this->edges) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
    }*/
    for (auto e : this->parens) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
    }
    for (int src : inputs) {
        for (auto ce : lib_gr->edges_from(src, GRAM_R)) {
            rec1[ce->get_target()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_from(src, GRAM_R0)) {
            rec1[ce->get_target()].push_back(ce);
        }
    }
    for (int tgt : outputs) {
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L)) {
            rec2[ce->get_source()].push_back(ce);
        }
        for (auto ce : lib_gr->edges_to(tgt, GRAM_L0)) {
            rec2[ce->get_source()].push_back(ce);
        }
    }
    auto real_start = clock();
    for (int n = 0; n < lib_nodes; n++) {
        if (rec1.count(n) != 0 && rec2.count(n) != 0) {
            auto &ls = rec1.at(n);
            auto &rs = rec2.at(n);
            for (auto ce1 : ls) {
                int src = ce1->get_source();
                auto premise = ce1->get_premise();
                for (auto ce2 : rs) {
                    int tgt = ce2->get_target();
                    auto new_premise = premise->merge(ce2->get_premise());
                    if (!lib_gr->is_redundant(src, tgt, GRAM_D, new_premise)) {
                        if (new_premise->size() > k) {
                            delete new_premise;
                            new_premise = new premise_1(edge::encode(edge(src, tgt, GRAM_D)));
                            if (!lib_gr->is_redundant(src, tgt, GRAM_D, new_premise)) {
                                auto new_edge = new conditional_edge(src, tgt, GRAM_D, new_premise);
                                lib_gr->add_record(new_edge);
                                lib_gr->add_edge(new_edge);
                                mapping->add2(src, tgt, GRAM_D, ce1, ce2);
                            } else {
                                mapping->add2(src, tgt, GRAM_D, ce1, ce2);
                            }
                        } else {
                            auto new_edge = new conditional_edge(src, tgt, GRAM_D, new_premise);
                            lib_gr->add_record(new_edge);
                            lib_gr->add_edge(new_edge);
                        }
                    } else {
                        delete new_premise;
                    }
                }
            }
        }
    }
    /*for (auto ce : lib_gr->all_parens()) {
        res->push_back(ce);
    }
    for (int src = 0; src < lib_gr->get_num_nodes(); src++) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : lib_gr->edges_from(src, s)) {
                res->push_back(ce);
            }
            auto &tmp = mapping->get_reocrd(src, s);
            for (auto it = tmp.begin(); it != tmp.end(); it++) {
                auto &dep = it->second;
                for (auto ce : dep) {
                    res->push_back(ce);
                }
            }
        }
    }*/

    auto start = clock();
    cerr << "On-the-fly D computation time: " << (start - real_start) * 1.0 / CLOCKS_PER_SEC << endl;

    vector<const conditional_edge*> *res = new vector<const conditional_edge*>();
    queue<size_t> work_list;
    unordered_set<size_t> vis;
    unordered_set<const conditional_edge*> hash;
    for (int src : inputs) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : lib_gr->edges_from(src, s)) {
                if (outputs.count(ce->get_target()) != 0) {
                    if (hash.count(ce) != 0) continue;
                    hash.insert(ce);
                    auto premise = ce->get_premise();
                    int sz = premise->size();
                    bool chk = true;
                    for (int i = 0; i < sz && chk; i++) {
                        auto fr = premise->get_one(i);
                        if ((fr >> 62) & 1) {
                            chk = (this->implemeted.count(fr) != 0);
                        }
                    }
                    if (chk) {
                        res->push_back(ce);
                        for (int i = 0; i < sz; i++) {
                            auto fr = premise->get_one(i);
                            if (!((fr >> 62) & 1) && !vis.count(fr)) {
                                vis.insert(fr);
                                work_list.push(fr);
                            }
                        }
                    }
                }
            }
        }
        /*for (auto ce : lib_gr->all_left_parens_from(src)) {
            if (outputs.count(ce->get_target()) != 0) {
                res->push_back(ce);
            }
        }
        for (auto ce : lib_gr->all_right_parens_from(src)) {
            if (outputs.count(ce->get_target()) != 0) {
                res->push_back(ce);
            }
        }*/
    }
    /*for (int tgt : outputs) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : lib_gr->edges_to(tgt, s)) {
                res->push_back(ce);
                auto premise = ce->get_premise();
                int sz = premise->size();
                for (int i = 0; i < sz; i++) {
                    work_list.push(premise->get_one(i));
                }
            }
        }
        for (auto ce : lib_gr->all_left_parens_to(tgt)) {
            res->push_back(ce);
        }
        for (auto ce : lib_gr->all_right_parens_to(tgt)) {
            res->push_back(ce);
        }
    }*/
    while (!work_list.empty()) {
        auto fr = work_list.front();
        work_list.pop();
        edge e = edge::decode(fr);
        if (mapping->has(e.get_source(), e.get_target(), e.get_symbol())) {
            for (auto ce : mapping->get(e.get_source(), e.get_target(), e.get_symbol())) {
                if (hash.count(ce) != 0) continue;
                hash.insert(ce);
                auto premise = ce->get_premise();
                int sz = premise->size();
                bool chk = true;
                for (int i = 0; i < sz && chk; i++) {
                    auto fr = premise->get_one(i);
                    if ((fr >> 62) & 1) {
                        chk = (this->implemeted.count(fr) != 0);
                    }
                }
                if (chk) {
                    res->push_back(ce);
                    for (int i = 0; i < sz; i++) {
                        auto fr = premise->get_one(i);
                        if (!((fr >> 62) & 1) && !vis.count(fr)) {
                            vis.insert(fr);
                            work_list.push(fr);
                        }
                    }
                }
            }
        }
    }

    auto end = clock();
    cerr << "Select time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    return res;
}

application_graph::application_graph(int num_nodes) {
    this->num_nodes = num_nodes;
    this->num_edges = 0;
    for (int s = 0; s < 7; s++) {
        this->forward[s] = new vector<const edge*>[num_nodes];
        this->backward[s] = new vector<const edge*>[num_nodes];
        this->record[s] = new unordered_set<int>[num_nodes];
        this->count[s] = 0;
    }
    this->forward_left = new vector<const edge*>[num_nodes];
    this->forward_right = new vector<const edge*>[num_nodes];
    this->backward_left = new vector<const edge*>[num_nodes];
    this->backward_right = new vector<const edge*>[num_nodes];
    this->rec_par = new unordered_map<int, unordered_set<int>>[num_nodes];
}

void application_graph::add_edge(const edge *e) {
    this->num_edges += 1;
    this->count[e->get_symbol()] += 1;
    this->forward[e->get_symbol()][e->get_source()].push_back(e);
    this->backward[e->get_symbol()][e->get_target()].push_back(e);
    this->edges.push_back(e);
}

void application_graph::add_record(const edge *e) {
    this->record[e->get_symbol()][e->get_source()].insert(e->get_target());
}

void application_graph::add_paren(const edge *e) {
    if (is_left_paren(e->get_symbol())) {
        this->forward_left[e->get_source()].push_back(e);
        this->backward_left[e->get_target()].push_back(e);
        this->left_parens[e->get_symbol()].push_back(e);
    } else {
        this->forward_right[e->get_source()].push_back(e);
        this->backward_right[e->get_target()].push_back(e);
        this->right_parens[e->get_symbol()].push_back(e);
    }
    this->parens.push_back(e);
    this->rec_par[e->get_source()][e->get_target()].insert(e->get_symbol());
}

bool application_graph::is_redundant(int source, int target, int symbol) const {
    return (this->record[symbol][source].count(target) != 0);
}

application_graph* application_graph::load(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes;
    fin >> num_nodes;
    application_graph *app_gr = new application_graph(num_nodes);
    string line;
    getline(fin, line);
    while (getline(fin, line)) {
        stringstream sin(line);
        int source, target;
        sin >> source >> target;
        string type;
        if (sin >> type) {
            int id;
            sin >> id;
            if (type[0] == 'L') {
                id += 10;
                app_gr->add_paren(new edge(source, target, id));
            } else if (type[0] == 'R') {
                id += 10;
                app_gr->add_paren(new edge(source, target, -id));
            } else {
                throw "bad implementation";
            }
        } else {
            app_gr->add_edge(new edge(source, target, GRAM_S0));
        }
    }
    return app_gr;
}

void application_graph::load_implemented(const string &file_name) {
    ifstream fin(file_name);
    for (size_t tmp; fin >> tmp; ) {
        tmp += 10;
        tmp |= 1UL << 62;
        this->implemeted.insert(tmp);
    }
    fin.close();
}

int application_graph::get_num_reach(int lib_nodes, Logger &logger) const {
    int res = 0;
    int cnt[7];
    memset(cnt, 0, sizeof(cnt));
    for (auto e : this->edges) {
        int sym = e->get_symbol();
        if (sym == GRAM_S0 || sym == GRAM_L0 || sym == GRAM_R0) {
            continue;
        }
        if (e->get_source() >= lib_nodes && e->get_target() >= lib_nodes) {
            //cerr << e->get_source() << " " << e->get_target() << " " << e->get_symbol() << endl;
            ++res;
            ++cnt[sym];
        }
    }
    cerr << "Client reachability: " << endl;
    cerr << "S: " << cnt[GRAM_S] << ", L: " << cnt[GRAM_L] << ", R: " << cnt[GRAM_R] << ", D: " << cnt[GRAM_D] << endl;
    cerr << "Total: " << res << endl;
    logger.log("#Reachability in Client: ");
    logger.log(res);
    logger.log("\n");
    return res;
}

bool application_graph::has_paren(int source, int target, int symbol) const {
    if (rec_par[source].count(target) != 0) {
        return (rec_par[source].at(target).count(symbol) != 0);
    } else {
        return false;
    }
}

application_graph* application_graph::load2(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes;
    fin >> num_nodes;
    application_graph *app_gr = new application_graph(num_nodes);
    string line;
    getline(fin, line);
    while (getline(fin, line)) {
        stringstream sin(line);
        int source, target;
        sin >> source >> target;
        string type;
        if (sin >> type) {
            int id;
            sin >> id;
            if (type[0] == 'L') {
                id += 10;
                if (type.length() > 1) {
                    app_gr->internals.insert(source);
                    app_gr->internals.insert(target);
                } else {
                    app_gr->add_paren(new edge(source, target, id));
                }
            } else if (type[0] == 'R') {
                id += 10;
                if (type.length() > 1) {
                    app_gr->internals.insert(source);
                    app_gr->internals.insert(target);
                } else {
                    app_gr->add_paren(new edge(source, target, -id));
                }
            } else if (type[0] == 'C') {
                app_gr->internals.insert(source);
                app_gr->internals.insert(target);
            } else {
                throw "bad implementation";
            }
        } else {
            app_gr->add_edge(new edge(source, target, GRAM_S0));
        }
    }
    return app_gr;
}

const unordered_set<int>& application_graph::get_internals() const {
    return this->internals;
}

void application_graph::set_internals(const unordered_set<int> &internals) {
    this->internals = internals;
}