//
// Created by Wayne on 16/1/9.
//

#ifndef REACHABILITY_APPLICATION_GRAPH_H
#define REACHABILITY_APPLICATION_GRAPH_H


#include <unordered_set>
#include <queue>
#include <sstream>
#include <cstring>
#include "conditional_edge.h"
#include "summarize_k.h"
#include "grammar.h"

using std::vector;
using std::unordered_set;

class application_graph {
    int num_nodes;
    int num_edges;
    vector<const edge*> edges;
    unordered_map<int, vector<const edge*>> left_parens, right_parens;
    vector<const edge*> parens;
    vector<const edge*> *forward_left, *forward_right, *backward_left, *backward_right;
    vector<const edge*> *forward[7], *backward[7];
    unordered_set<int> *record[7];
    unordered_map<int, unordered_set<int>> *rec_par;
    size_t count[7];
    unordered_set<size_t> implemeted;
    unordered_set<int> internals;
public:
    const unordered_map<int, vector<const edge*>>& get_left_parens() const { return this->left_parens; }
    const unordered_map<int, vector<const edge*>>& get_right_parens() const { return this->right_parens; };
    int get_num_nodes() const { return this->num_nodes; }
    int get_num_edges() const { return this->num_edges; }
    const vector<const edge*>& all_edges() const { return this->edges; }
    const vector<const edge*>& all_parens() const { return this->parens; }
    const vector<const edge*>& edges_from(int source, int symbol) const { return this->forward[symbol][source]; }
    const vector<const edge*>& edges_to(int target, int symbol) const { return this->backward[symbol][target]; }
    application_graph(int num_nodes);
    void add_record(const edge *e);
    void add_edge(const edge *e);
    void add_paren(const edge *e);
    const vector<const edge*>& all_left_parens_from(int source) const { return this->forward_left[source]; };
    const vector<const edge*>& all_right_parens_from(int source) const { return this->forward_right[source]; };
    const vector<const edge*>& all_left_parens_to(int target) const { return this->backward_left[target]; }
    const vector<const edge*>& all_right_parens_to(int target) const { return this->backward_right[target]; }
    bool is_redundant(int source, int target, int symbol) const;
    const vector<const conditional_edge*>* get_conds_f(library_graph *lib_gr) const;
    const vector<const conditional_edge*>* get_conds_k(int k, library_graph *lib_gr, hocra_mapping *mapping) const;
    const vector<const conditional_edge*>* get_conds_cbcla(library_graph *lib_gr) const;
    const unordered_set<size_t>& get_implemented() const { return this->implemeted; }
    void load_implemented(const string &file_name);
    int get_num_reach(int lib_nodes, Logger &logger) const;
    bool has_paren(int source, int target, int symbol) const;
    const unordered_set<int>& get_internals() const;
    void set_internals(const unordered_set<int> &internals);
public:
    static application_graph* load(const string &file_name);
    static application_graph* load2(const string &file_name);
};


#endif //REACHABILITY_APPLICATION_GRAPH_H
