//
// Created by Wayne on 3/17/16.
//

#include <iostream>
#include <sys/time.h>
#include <sys/resource.h>
#include "library_graph.h"
#include "summarize_k.h"
#include "application_graph.h"
#include "analyze.h"
#include "Logger.h"

using namespace std;

static long get_memory_usage() {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    return usage.ru_maxrss;
}

/*
 * argv[1] : k
 * argv[2] : LibraryGraph.dat
 * argv[3] : benchmark.txt
 * argv[4] : evaluation.log
 */
int main(int argc, char *argv[]) {
    ios::sync_with_stdio(false);

    Logger logger(argv[4]);

    cerr << "Loading ..." << endl;
    const library_graph *lib_gr = library_graph::load2(argv[2]);
    cerr << "Load done. Node: " << lib_gr->get_num_nodes() << ", Edge: " << lib_gr->get_num_edges() << ", Paren: " << lib_gr->all_parens().size() << endl;
    int k = atoi(argv[1]);
    cerr << "k: " << k << endl;

    logger.log("Evaluation on ConCRA-");
    logger.log(k);
    logger.log("\n");
    logger.log("======================\n");
    logger.log("#Library Node: ");
    logger.log(lib_gr->get_num_nodes());
    logger.log("\n");
    logger.log("#Library Edge: ");
    logger.log(lib_gr->get_num_edges());
    logger.log("\n");

    cerr << "Summarizing ..." << endl;
    pair<library_graph*, hocra_mapping*> tmp = analyze_libraries_k(k, lib_gr, lib_gr->extract_pairs2(), logger);
    cerr << "Summarize done." << endl;

    long lib_mem = get_memory_usage();
    cerr << "Summarize memory: " << lib_mem / 1024.0 << endl;

    logger.log("Memory of Summarization: ");
    logger.log(lib_mem / 1024.0);
    logger.log("\n");

    library_graph *tmp_gr = tmp.first;
    hocra_mapping *mapping = tmp.second;

    ifstream benchmark(argv[3]);
    for (string subject_dat, implemented_dat; benchmark >> subject_dat >> implemented_dat; ) {
        cerr << "Loading " << subject_dat << " ..." << endl;
        application_graph *app_gr = application_graph::load(subject_dat);
        app_gr->load_implemented(implemented_dat);
        cerr << "Load done." << endl;

        logger.log("----\n");
        logger.log("Subject: ");
        logger.log(subject_dat);
        logger.log("\n");

        cerr << "Analyzing " << subject_dat << " ..." << endl;
        const application_graph *res_gr = analyze_hocra_k(k, app_gr, tmp_gr, mapping, logger);
        cerr << "Analyze done." << endl;

        long app_mem = get_memory_usage() - lib_mem;
        lib_mem = get_memory_usage();
        cerr << "Analyze memory: " << app_mem / 1024.0 << endl;

        logger.log("Memory of Analysis: ");
        logger.log(app_mem / 1024.0);
        logger.log("\n");

        res_gr->get_num_reach(lib_gr->get_num_nodes(), logger);

        logger.log("----\n");
    }

    logger.close();

    return 0;
}