//
// Created by Wayne on 15/12/29.
//

#include "library_graph.h"

using std::ifstream;
using std::getline;
using std::unordered_multimap;
using std::make_pair;
using std::stringstream;
using std::cerr;
using std::endl;

library_graph::library_graph(int num_nodes) {
    this->num_nodes = num_nodes;
    //this->num_edges = 0;
    for (int s = 0; s < 7; s++) {
        this->forward[s] = new vector<const conditional_edge*>[num_nodes];
        this->backward[s] = new vector<const conditional_edge*>[num_nodes];
    }
    this->signatures.resize(num_nodes);
    for (int s = 0; s < 7; s++) {
        this->record[s] = new unordered_map<int, unordered_map<size_t, vector<const conditional_edge*>>>[num_nodes];
        this->count[s] = 0;
    }
    this->forward_left = new vector<const conditional_edge*>[num_nodes];
    this->forward_right = new vector<const conditional_edge*>[num_nodes];
    this->backward_left = new vector<const conditional_edge*>[num_nodes];
    this->backward_right = new vector<const conditional_edge*>[num_nodes];
}

void library_graph::add_record(const conditional_edge *ce) {
    auto premise = ce->get_premise();
    premise->add2(this->record[ce->get_symbol()][ce->get_source()][ce->get_target()], ce);
}

void library_graph::add_edge(const conditional_edge *ce) {
    //this->num_edges += 1;
    this->count[ce->get_symbol()] += 1;
    this->forward[ce->get_symbol()][ce->get_source()].push_back(ce);
    this->backward[ce->get_symbol()][ce->get_target()].push_back(ce);
}

bool library_graph::is_redundant(int source, int target, int symbol, premise_base *premise) const {
    /*if (this->record[symbol][source].count(target) != 0) {
        const auto &all = this->record[symbol][source].find(target)->second;
        if (all.count(1UL << 63) != 0) {
            return true;
        }
        int sz = premise->size();
        for (int id = 0; id < sz; id++) {
            auto one = premise->get_one(id);
            if (all.count(one) != 0) {
                const auto &edges = all.at(one);
                for (auto ce : edges) {
                    if (ce->get_premise()->is_covered_by(premise)) {
                        return true;
                    }
                }
            }
        }
    }
    return false;*/
    if (this->record[symbol][source].count(target) != 0) {
        const auto &all = this->record[symbol][source].find(target)->second;
        vector<size_t> subsets = premise->subsets();
        for (size_t hash : subsets) {
            if (all.count(hash) != 0) {
                const auto &edges = all.at(hash);
                for (auto ce : edges) {
                    if (ce->get_premise()->is_covered_by(premise)) {
                        return true;
                    }
                }
            }
        }
    }
    return false;
}

const library_graph* library_graph::load(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes, num_edges;
    fin >> num_nodes >> num_edges;
    library_graph *lib_gr = new library_graph(num_nodes);
    for (int n = 0; n < num_nodes; n++) {
        int type, index;
        fin >> type >> index;
        string method_str, unit_str;
        getline(fin, method_str);
        if (method_str.empty() || method_str[0] == '\n') {
            getline(fin, method_str);
        }
        getline(fin, unit_str);
        lib_gr->signatures[n] = method_str + ":" + unit_str;
        switch (type) {
            case INPUT_NODE:
                lib_gr->bound_inputs.insert(n);
                break;
            case RETURN_NODE:
                lib_gr->bound_returns.insert(n);
                break;
            case CALL_NODE:
                lib_gr->bound_calls.insert(n);
                break;
            case OUTPUT_NODE:
                lib_gr->bound_outputs.insert(n);
                break;
            default:
                throw "bad implementation";
        }
    }
    for (int m = 0; m < num_edges; m++) {
        int source, target, symbol;
        fin >> source >> target >> symbol;
        if (symbol == GRAM_S) symbol = GRAM_S0;
        conditional_edge *ce = new conditional_edge(source, target, symbol, NULL);
        if (is_paren(symbol)) {
            lib_gr->add_paren(ce);
        } else {
            lib_gr->add_edge(ce);
        }
    }
    fin.close();
    return lib_gr;
}

const vector<const conditional_edge*>& library_graph::edges_from(int source, int symbol) const {
    return this->forward[symbol][source];
}

const vector<const conditional_edge*>& library_graph::edges_to(int target, int symbol) const {
    return this->backward[symbol][target];
}

const vector<pair<int, int>>* library_graph::extract_pairs() const {
    vector<pair<int, int>> *res = new vector<pair<int, int>>();
    unordered_multimap<string, int> sig2site;
    for (int n : this->bound_calls) {
        sig2site.insert(make_pair(this->signatures[n], n));
    }
    for (int n : this->bound_returns) {
        auto &signature = this->signatures[n];
        for (auto it = sig2site.find(signature); it != sig2site.end() && it->first == signature; it++) {
            res->push_back(make_pair(it->second, n));
        }
    }
    return res;
}

const library_graph* library_graph::load2(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes;
    fin >> num_nodes;
    library_graph *lib_gr = new library_graph(num_nodes);
    string line;
    getline(fin, line);
    while (getline(fin, line)) {
        stringstream sin(line);
        int source, target;
        sin >> source >> target;
        string type;
        if (sin >> type) {
            int id;
            sin >> id;
            if (type[0] == 'C') {
                lib_gr->callbacks.push_back(make_pair(source, target));
                //lib_gr->bound_calls.insert(source);
                //lib_gr->bound_returns.insert(target);
            } else if (type[0] == 'L') {
                id += 10;
                if (type.length() > 1) {
                    premise_base *new_premise = new premise_1(1UL << 62 | id);
                    lib_gr->add_paren(new conditional_edge(source, target, id, new_premise));
                } else {
                    lib_gr->add_paren(new conditional_edge(source, target, id, new premise_0));
                }
            } else if (type[0] == 'R') {
                id += 10;
                if (type.length() > 1) {
                    premise_base *new_premise = new premise_1(1UL << 62 | id);
                    lib_gr->add_paren(new conditional_edge(source, target, -id, new_premise));
                } else {
                    lib_gr->add_paren(new conditional_edge(source, target, -id, new premise_0));
                }
            } else {
                throw "bad implementation";
            }
        } else {
            lib_gr->add_edge(new conditional_edge(source, target, GRAM_S0, new premise_0));
        }
    }
    return lib_gr;
}

const vector<pair<int, int>>* library_graph::extract_pairs2() const {
    return &(this->callbacks);
}

void library_graph::add_paren(const conditional_edge *ce) {
    if (is_left_paren(ce->get_symbol())) {
        this->forward_left[ce->get_source()].push_back(ce);
        this->backward_left[ce->get_target()].push_back(ce);
    } else {
        this->forward_right[ce->get_source()].push_back(ce);
        this->backward_right[ce->get_target()].push_back(ce);
    }
    this->parens.push_back(ce);
}


void library_graph::evaluate() const {
    unordered_map<int, int> cnt;
    for (int src = 0; src < this->num_nodes; src++) {
        for (int s = 0; s < 7; s++) {
            for (auto ce : this->edges_from(src, s)) {
                cnt[ce->get_premise()->size()] += 1;
            }
        }
    }
    for (auto &p : cnt) {
        cerr << "Premise size " << p.first << ": " << p.second << endl;
    }
}

const library_graph* library_graph::load3(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes;
    fin >> num_nodes;
    library_graph *lib_gr = new library_graph(num_nodes);
    string line;
    getline(fin, line);
    while (getline(fin, line)) {
        stringstream sin(line);
        int source, target;
        sin >> source >> target;
        string type;
        if (sin >> type) {
            int id;
            sin >> id;
            if (type[0] == 'L') {
                id += 10;
                if (type.length() > 1) {

                } else {
                    lib_gr->add_paren(new conditional_edge(source, target, id, new premise_0));
                }
            } else if (type[0] == 'R') {
                id += 10;
                if (type.length() > 1) {

                } else {
                    lib_gr->add_paren(new conditional_edge(source, target, -id, new premise_0));
                }
            } else if (type[0] == 'C') {
                lib_gr->callbacks.push_back(make_pair(source, target));
            } else {
                throw "bad implementation";
            }
        } else {
            lib_gr->add_edge(new conditional_edge(source, target, GRAM_S0, new premise_0));
        }
    }
    return lib_gr;
}