//
// Created by Wayne on 15/12/29.
//

#ifndef REACHABILITY_LIBRARY_GRAPH_H
#define REACHABILITY_LIBRARY_GRAPH_H


#include <vector>
#include <string>
#include <fstream>
#include <unordered_map>
#include <climits>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <unordered_set>
#include "conditional_edge.h"
#include "edge.h"
#include "grammar.h"

using std::vector;
using std::string;
using std::pair;
using std::unordered_map;
using std::unordered_set;

class library_graph {
    int num_nodes;
    //int num_edges;
    vector<const conditional_edge*> parens;
    vector<const conditional_edge*> *forward_left, *forward_right, *backward_left, *backward_right;
    vector<const conditional_edge*> *forward[7], *backward[7];
    unordered_map<int, unordered_map<size_t, vector<const conditional_edge*>>> *record[7];
    size_t count[7];
    unordered_set<int> bound_inputs, bound_returns, bound_calls, bound_outputs;
    vector<string> signatures;
    vector<pair<int, int>> callbacks;
public:
    const size_t get_count(int symbol) const { return this->count[symbol]; }
    void add_bound_call(int n) { this->bound_calls.insert(n); }
    void add_bound_return(int n) { this->bound_returns.insert(n); }
    const unordered_set<int>& get_bound_calls() const { return this->bound_calls; }
    const unordered_set<int>& get_bound_returns() const { return this->bound_returns; }
    int get_num_nodes() const { return num_nodes; }
    int get_num_edges() const {
        int res = 0;
        for (int s = 0; s < 7; s++) {
            res += this->count[s];
        }
        return res;
    }
    const vector<const conditional_edge*>& all_parens() const { return parens; }
    const vector<const conditional_edge*>& edges_from(int source, int symbol) const;
    const vector<const conditional_edge*>& edges_to(int target, int symbol) const;
    library_graph(int num_nodes);
    void add_record(const conditional_edge *ce);
    void add_edge(const conditional_edge *ce);
    void add_paren(const conditional_edge *ce);
    const vector<const conditional_edge*>& all_left_parens_from(int source) const { return forward_left[source]; }
    const vector<const conditional_edge*>& all_right_parens_from(int source) const { return forward_right[source]; }
    const vector<const conditional_edge*>& all_left_parens_to(int target) const { return backward_left[target]; }
    const vector<const conditional_edge*>& all_right_parens_to(int target) const { return backward_right[target]; }
    bool is_redundant(int source, int target, int symbol, premise_base *premise) const;
    bool is_bound_return(int n) const { return (bound_returns.count(n) != 0); }
    bool is_bound_call(int n) const { return (bound_calls.count(n) != 0); }
    bool is_bound_input(int n) const { return (bound_inputs.count(n) != 0); }
    bool is_bound_output(int n) const { return (bound_outputs.count(n) != 0); }
    const vector<pair<int, int>>* extract_pairs() const;
    const vector<pair<int, int>>* extract_pairs2() const;
    void evaluate() const;
public:
    static const library_graph* load(const string &file_name);
    static const library_graph* load2(const string &file_name);
    static const library_graph* load3(const string &file_name);
public:
    static const int INPUT_NODE = 0, RETURN_NODE = 1, CALL_NODE = 2, OUTPUT_NODE = 3;
};


#endif //REACHABILITY_LIBRARY_GRAPH_H
