//
// Created by Wayne on 15/12/30.
//

#include "summarize_f.h"

using std::cerr;
using std::endl;
using std::make_pair;
using std::map;

static inline void add(int symbol, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    /*if (src == tgt) {
        if (!ref) delete premise;
        return;
    }*/
    if (!gr->is_redundant(src, tgt, symbol, premise)) {
        auto new_edge = new conditional_edge(src, tgt, symbol, premise);
        wl.push(premise->size(), new_edge);
        gr->add_record(new_edge);
    } else {
        if (!ref) delete premise;
    }
}

static inline void addS(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_S, wl, gr, src, tgt, premise, ref);
}

static inline void addL(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_L, wl, gr, src, tgt, premise, ref);
}

static inline void addR(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_R, wl, gr, src, tgt, premise, ref);
}

static inline void addD(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_D, wl, gr, src, tgt, premise, ref);
}

static inline void addS0(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_S0, wl, gr, src, tgt, premise, ref);
}

static inline void addL0(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_L0, wl, gr, src, tgt, premise, ref);
}

static inline void addR0(multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    add(GRAM_R0, wl, gr, src, tgt, premise, ref);
}

#define FORWARD(X) { \
        premise_base *new_premise = premise->merge(forward->get_premise()); \
        int tgt = forward->get_target(); \
        add##X(work_list, tmp_gr, source, tgt, new_premise); \
}
#define BACKWARD(X) { \
        premise_base *new_premise = premise->merge(backward->get_premise()); \
        int src = backward->get_source(); \
        add##X(work_list, tmp_gr, src, target, new_premise); \
}

library_graph* analyze_libraries_f(const library_graph *lib_gr, const vector<pair<int, int>> *conds, Logger &logger) {
    auto start = clock();

    multilevel_queue<const conditional_edge*> work_list;
    library_graph *tmp_gr = new library_graph(lib_gr->get_num_nodes());

    unordered_map<int, vector<const conditional_edge*>> left_parens;
    vector<map<int, vector<pair<premise_base*, pair<int, int>>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    for (auto ce : lib_gr->all_parens()) {
        tmp_gr->add_paren(ce);
        if (is_left_paren(ce->get_symbol())) {
            left_parens[ce->get_symbol()].push_back(ce);
        }
    }
    for (auto ce : lib_gr->all_parens()) {
        if (is_right_paren(ce->get_symbol())) {
            int left_paren = -ce->get_symbol();
            if (left_parens.count(left_paren) != 0) {
                for (auto ce2 : left_parens.find(left_paren)->second) {
                    record[ce2->get_target()][ce->get_source()].push_back(make_pair(ce2->get_premise()->merge(ce->get_premise()), make_pair(ce2->get_source(), ce->get_target())));
                }
            }
        }
    }

    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto ce : lib_gr->edges_from(src, GRAM_S0)) {
            addS0(work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        }
    }
    for (auto ce : lib_gr->all_parens()) {
        if (is_left_paren(ce->get_symbol())) {
            addL0(work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else if (is_right_paren(ce->get_symbol())) {
            addR0(work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else {
            throw "bad implementation";
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                auto new_premise = p.first;
                auto &q = p.second;
                addS0(work_list, tmp_gr, q.first, q.second, new_premise, true);
            }
        }
    }
    for (auto it = conds->begin(); it != conds->end(); it++) {
        int source = it->first;
        int target = it->second;
        edge base_edge(source, target, GRAM_S0);
        addS0(work_list, tmp_gr, source, target, new premise_1(edge::encode(base_edge)));
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);

        int source = ce->get_source();
        int target = ce->get_target();
        int symbol = ce->get_symbol();
        auto premise = ce->get_premise();

        assert(!is_paren(symbol));

        if (symbol == GRAM_S0) {
            // S -> @S0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }
            // S -> S@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(S);
            }
            // S -> @S0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(S);
            }*/
            // L -> L@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                BACKWARD(L);
            }
            // L -> @S0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                FORWARD(L);
            }*/
            // L -> @S0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }
            // R -> R@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                BACKWARD(R);
            }
            // R -> @S0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                FORWARD(R);
            }*/
            // R -> @S0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }
            // S0 -> {i@S0}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first;
                    int tgt = q.second;
                    addS0(work_list, tmp_gr, src, tgt, new_premise);
                }
            }
        } else if (symbol == GRAM_L0) {
            // L -> @L0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }
            // L -> L@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                BACKWARD(L);
            }
            // L -> @L0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                FORWARD(L);
            }*/
            // L -> S@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(L);
            }
            // L -> @L0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(L);
            }*/
            // L -> S0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }
            // L -> @L0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
        } else if (symbol == GRAM_R0) {
            // R -> @R0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }
            // R -> R@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                BACKWARD(R);
            }
            // R -> @R0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                FORWARD(R);
            }*/
            // R -> S@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(R);
            }
            // R -> @R0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(R);
            }*/
            // R -> S0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }
            // R -> @R0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
        } else if (symbol == GRAM_S) {
            // S -> @SS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }*/
            // L -> @SL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }*/
            // R -> @SR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }*/
            // S0 -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first;
                    int tgt = q.second;
                    addS0(work_list, tmp_gr, src, tgt, new_premise);
                }
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }*/
            // L -> @LS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
            // L -> S0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }*/
        } else if (symbol == GRAM_R) {
            // R -> @RR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }*/
            // R -> @RS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
            // R -> S0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }*/
        } else {
            throw "bad implementation";
        }
    }

    /*for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                addD(work_list, tmp_gr, src, tgt, new_premise);
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                addD(work_list, tmp_gr, src, tgt, new_premise);
            }
        }
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R0)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                addD(work_list, tmp_gr, src, tgt, new_premise);
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                addD(work_list, tmp_gr, src, tgt, new_premise);
            }
        }
    }
    while (!work_list.empty()) {
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);
    }*/

    cerr << "Total edges: " << tmp_gr->get_num_edges() << endl;

    auto end = clock();
    auto consumed_time = (end - start) * 1.0 / CLOCKS_PER_SEC;
    cerr << "Summarize time: " << consumed_time << endl;

    logger.log("#Edge in the Summary: ");
    logger.log(tmp_gr->get_num_edges());
    logger.log("\n");
    logger.log("Time of Summarization: ");
    logger.log(consumed_time);
    logger.log("\n");

    return tmp_gr;
}