//
// Created by Wayne on 16/1/5.
//

#include "summarize_k.h"

using std::map;
using std::make_pair;
using std::cerr;
using std::endl;
using std::thread;

// 0 - not add in
// 1 - add in
// 2 - add in with limit

//static size_t new_conds_cnt = 0;

static inline int add(int symbol, int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    /*if (src == tgt) {
        if (!ref) delete premise;
        return 0;
    }*/
    if (!gr->is_redundant(src, tgt, symbol, premise)) {
        int res = 1;
        if (premise->size() > k) {
            res = 2;
            if (!ref) delete premise;
            premise = new premise_1(edge::encode(edge(src, tgt, symbol)));
            if (gr->is_redundant(src, tgt, symbol, premise)) {
                delete premise;
                return 2;
            }
        }
        //if (res == 2) new_conds_cnt += 1;
        auto new_edge = new conditional_edge(src, tgt, symbol, premise);
        wl.push(premise->size(), new_edge);
        gr->add_record(new_edge);
        return res;
    } else {
        if (!ref) delete premise;
        return 0;
    }
}

static inline int addS(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_S, k, wl, gr, src, tgt, premise, ref);
}

static inline int addL(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_L, k, wl, gr, src, tgt, premise, ref);
}

static inline int addR(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_R, k, wl, gr, src, tgt, premise, ref);
}

static inline int addD(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_D, k, wl, gr, src, tgt, premise, ref);
}

static inline int addS0(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_S0, k, wl, gr, src, tgt, premise, ref);
}

static inline int addL0(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_L0, k, wl, gr, src, tgt, premise, ref);
}

static inline int addR0(int k, multilevel_queue<const conditional_edge*> &wl, library_graph *gr, int src, int tgt, premise_base *premise, bool ref = false) {
    return add(GRAM_R0, k, wl, gr, src, tgt, premise, ref);
}

#define FORWARD(X) { \
        premise_base *new_premise = premise->merge(forward->get_premise()); \
        int tgt = forward->get_target(); \
        if (add##X(k, work_list, tmp_gr, source, tgt, new_premise) == 2) { \
            mapping->add2(source, tgt, GRAM_##X, ce, forward); \
        } \
}
#define BACKWARD(X) { \
        premise_base *new_premise = premise->merge(backward->get_premise()); \
        int src = backward->get_source(); \
        if (add##X(k, work_list, tmp_gr, src, target, new_premise) == 2) { \
            mapping->add2(src, target, GRAM_##X, backward, ce); \
        } \
}

pair<library_graph*, hocra_mapping*> analyze_libraries_k(int k, const library_graph *lib_gr, const vector<pair<int, int>> *conds, Logger &logger) {
    auto start = clock();

    multilevel_queue<const conditional_edge*> work_list;
    library_graph *tmp_gr = new library_graph(lib_gr->get_num_nodes());
    hocra_mapping *mapping = new hocra_mapping(lib_gr->get_num_nodes());

    unordered_map<int, vector<const conditional_edge*>> left_parens;
    vector<map<int, vector<pair<premise_base*, pair<const conditional_edge*, const conditional_edge*>>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    for (auto ce : lib_gr->all_parens()) {
        tmp_gr->add_paren(ce);
        if (is_left_paren(ce->get_symbol())) {
            left_parens[ce->get_symbol()].push_back(ce);
        }
    }
    for (auto ce : lib_gr->all_parens()) {
        if (is_right_paren(ce->get_symbol())) {
            int left_paren = -ce->get_symbol();
            if (left_parens.count(left_paren) != 0) {
                for (auto ce2 : left_parens.find(left_paren)->second) {
                    record[ce2->get_target()][ce->get_source()].push_back(make_pair(ce2->get_premise()->merge(ce->get_premise()), make_pair(ce2, ce)));
                }
            }
        }
    }

    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto ce : lib_gr->edges_from(src, GRAM_S0)) {
            addS0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        }
    }
    for (auto ce : lib_gr->all_parens()) {
        if (is_left_paren(ce->get_symbol())) {
            addL0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else if (is_right_paren(ce->get_symbol())) {
            addR0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else {
            throw "bad implementation";
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                auto new_premise = p.first;
                auto &q = p.second;
                if (addS0(k, work_list, tmp_gr, q.first->get_source(), q.second->get_target(), new_premise, true) == 2) {
                    mapping->add2(q.first->get_source(), q.second->get_target(), GRAM_S0, q.first, q.second);
                }
            }
        }
    }
    for (auto it = conds->begin(); it != conds->end(); it++) {
        int source = it->first;
        int target = it->second;
        edge base_edge(source, target, GRAM_S0);
        addS0(k, work_list, tmp_gr, source, target, new premise_1(edge::encode(base_edge)));
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);

        int source = ce->get_source();
        int target = ce->get_target();
        int symbol = ce->get_symbol();
        auto premise = ce->get_premise();

        assert(!is_paren(symbol));

        if (symbol == GRAM_S0) {
            // S -> @S0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }
            // S -> S@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(S);
            }
            // S -> @S0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(S);
            }*/
            // L -> L@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                BACKWARD(L);
            }
            // L -> @S0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                FORWARD(L);
            }*/
            // L -> @S0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }
            // R -> R@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                BACKWARD(R);
            }
            // R -> @S0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                FORWARD(R);
            }*/
            // R -> @S0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }
            // S0 -> {i@S0}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first->get_source();
                    int tgt = q.second->get_target();
                    if (addS0(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                        mapping->add3(src, tgt, GRAM_S0, q.first, ce, q.second);
                    }
                }
            }
        } else if (symbol == GRAM_L0) {
            // L -> @L0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }
            // L -> L@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                BACKWARD(L);
            }
            // L -> @L0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                FORWARD(L);
            }*/
            // L -> S@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(L);
            }
            // L -> @L0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(L);
            }*/
            // L -> S0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }
            // L -> @L0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
        } else if (symbol == GRAM_R0) {
            // R -> @R0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }
            // R -> R@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                BACKWARD(R);
            }
            // R -> @R0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                FORWARD(R);
            }*/
            // R -> S@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(R);
            }
            // R -> @R0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(R);
            }*/
            // R -> S0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }
            // R -> @R0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
        } else if (symbol == GRAM_S) {
            // S -> @SS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }*/
            // L -> @SL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }*/
            // R -> @SR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }*/
            // S0 -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first->get_source();
                    int tgt = q.second->get_target();
                    if (addS0(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                        mapping->add3(src, tgt, GRAM_S0, q.first, ce, q.second);
                    }
                }
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }*/
            // L -> @LS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
            // L -> S0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }*/
        } else if (symbol == GRAM_R) {
            // R -> @RR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }*/
            // R -> @RS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
            // R -> S0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }*/
        } else {
            throw "bad implementation";
        }
    }

    /*for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
        }
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R0)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
        }
    }
    while (!work_list.empty()) {
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);
    }*/

    cerr << "Total edges: " << tmp_gr->get_num_edges() << endl;
    //cerr << "New conditions: " << new_conds_cnt << endl;

    auto end = clock();
    auto consumed_time = (end - start) * 1.0 / CLOCKS_PER_SEC;
    cerr << "Summarize time: " << consumed_time << endl;

    logger.log("#Edge in the Summary: ");
    logger.log(tmp_gr->get_num_edges());
    logger.log("\n");
    logger.log("Time of Summarization: ");
    logger.log(consumed_time);
    logger.log("\n");

    return make_pair(tmp_gr, mapping);
}

struct {
    int k;
    const library_graph *lib_gr;
    library_graph *tmp_gr;
    hocra_mapping *mapping;
} global_attr;

static void calc_l() {
    multilevel_queue<const conditional_edge*> work_list;
    library_graph *tmp_gr = global_attr.tmp_gr;
    hocra_mapping *mapping = global_attr.mapping;
    int k = global_attr.k;

    for (auto ce : global_attr.lib_gr->all_parens()) {
        if (is_left_paren(ce->get_symbol())) {
            addL0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        }
    }

    while (!work_list.empty()) {
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);

        int source = ce->get_source();
        int target = ce->get_target();
        int symbol = ce->get_symbol();
        auto premise = ce->get_premise();

        assert(!is_paren(symbol));

        if (symbol == GRAM_L0) {
            // L -> @L0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }
            // L -> L@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                BACKWARD(L);
            }
            // L -> @L0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                FORWARD(L);
            }*/
            // L -> S@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(L);
            }
            // L -> @L0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(L);
            }*/
            // L -> S0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }
            // L -> @L0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                FORWARD(L);
            }
            // L -> L0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                BACKWARD(L);
            }*/
            // L -> @LS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(L);
            }
            // L -> S0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(L);
            }*/
        } else {
            throw "bad implementation";
        }
    }
}

static void calc_r() {
    multilevel_queue<const conditional_edge*> work_list;
    library_graph *tmp_gr = global_attr.tmp_gr;
    hocra_mapping *mapping = global_attr.mapping;
    int k = global_attr.k;

    for (auto ce : global_attr.lib_gr->all_parens()) {
        if (is_right_paren(ce->get_symbol())) {
            addR0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        }
    }

    while (!work_list.empty()) {
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);

        int source = ce->get_source();
        int target = ce->get_target();
        int symbol = ce->get_symbol();
        auto premise = ce->get_premise();

        assert(!is_paren(symbol));

        if (symbol == GRAM_R0) {
            // R -> @R0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }
            // R -> R@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                BACKWARD(R);
            }
            // R -> @R0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                FORWARD(R);
            }*/
            // R -> S@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(R);
            }
            // R -> @R0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(R);
            }*/
            // R -> S0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }
            // R -> @R0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
        } else if (symbol == GRAM_R) {
            // R -> @RR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                FORWARD(R);
            }
            // R -> R0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                BACKWARD(R);
            }*/
            // R -> @RS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(R);
            }
            // R -> S0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(R);
            }*/
        } else {
            throw "bad implementation";
        }
    }
}

pair<library_graph*, hocra_mapping*> analyze_libraries_k_multi(int k, const library_graph *lib_gr, const vector<pair<int, int>> *conds) {
    auto start = clock();

    multilevel_queue<const conditional_edge*> work_list;
    library_graph *tmp_gr = new library_graph(lib_gr->get_num_nodes());
    hocra_mapping *mapping = new hocra_mapping(lib_gr->get_num_nodes());

    unordered_map<int, vector<const conditional_edge*>> left_parens;
    vector<map<int, vector<pair<premise_base*, pair<const conditional_edge*, const conditional_edge*>>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    for (auto ce : lib_gr->all_parens()) {
        tmp_gr->add_paren(ce);
        if (is_left_paren(ce->get_symbol())) {
            left_parens[ce->get_symbol()].push_back(ce);
        }
    }
    for (auto ce : lib_gr->all_parens()) {
        if (is_right_paren(ce->get_symbol())) {
            int left_paren = -ce->get_symbol();
            if (left_parens.count(left_paren) != 0) {
                for (auto ce2 : left_parens.find(left_paren)->second) {
                    record[ce2->get_target()][ce->get_source()].push_back(make_pair(ce2->get_premise()->merge(ce->get_premise()), make_pair(ce2, ce)));
                }
            }
        }
    }

    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto ce : lib_gr->edges_from(src, GRAM_S0)) {
            addS0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        }
    }
    /*for (auto ce : lib_gr->all_parens()) {
        if (is_left_paren(ce->get_symbol())) {
            addL0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else if (is_right_paren(ce->get_symbol())) {
            addR0(k, work_list, tmp_gr, ce->get_source(), ce->get_target(), ce->get_premise(), true);
        } else {
            throw "bad implementation";
        }
    }*/
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                auto new_premise = p.first;
                auto &q = p.second;
                if (addS0(k, work_list, tmp_gr, q.first->get_source(), q.second->get_target(), new_premise, true) == 2) {
                    mapping->add2(q.first->get_source(), q.second->get_target(), GRAM_S0, q.first, q.second);
                }
            }
        }
    }
    for (auto it = conds->begin(); it != conds->end(); it++) {
        int source = it->first;
        int target = it->second;
        edge base_edge(source, target, GRAM_S0);
        addS0(k, work_list, tmp_gr, source, target, new premise_1(edge::encode(base_edge)));
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);

        int source = ce->get_source();
        int target = ce->get_target();
        int symbol = ce->get_symbol();
        auto premise = ce->get_premise();

        assert(!is_paren(symbol));

        if (symbol == GRAM_S0) {
            // S -> @S0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }
            // S -> S@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                BACKWARD(S);
            }
            // S -> @S0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                FORWARD(S);
            }*/
            // S0 -> {i@S0}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first->get_source();
                    int tgt = q.second->get_target();
                    if (addS0(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                        mapping->add3(src, tgt, GRAM_S0, q.first, ce, q.second);
                    }
                }
            }
        } else if (symbol == GRAM_S) {
            // S -> @SS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                FORWARD(S);
            }
            // S -> S0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                BACKWARD(S);
            }*/
            // S0 -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    auto new_premise = premise->merge(p.first);
                    auto &q = p.second;
                    int src = q.first->get_source();
                    int tgt = q.second->get_target();
                    if (addS0(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                        mapping->add3(src, tgt, GRAM_S0, q.first, ce, q.second);
                    }
                }
            }
        } else {
            throw "bad implementation";
        }
    }

    global_attr.k = k;
    global_attr.lib_gr = lib_gr;
    global_attr.tmp_gr = tmp_gr;
    global_attr.mapping = mapping;

    thread thread_l(calc_l);
    thread thread_r(calc_r);
    thread_l.join();
    thread_r.join();

    /*for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
        }
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R0)) {
            int src = e1->get_source();
            auto premise = e1->get_premise();
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L0)) {
                int tgt = e2->get_target();
                auto new_premise = premise->merge(e2->get_premise());
                if (addD(k, work_list, tmp_gr, src, tgt, new_premise) == 2) {
                    mapping->add2(src, tgt, GRAM_D, e1, e2);
                }
            }
        }
    }
    while (!work_list.empty()) {
        auto ce = work_list.pop();
        tmp_gr->add_edge(ce);
    }*/

    cerr << "Total edges: " << tmp_gr->get_num_edges() << endl;
    //cerr << "New conditions: " << new_conds_cnt << endl;

    auto end = clock();
    cerr << "Summarize time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    return make_pair(tmp_gr, mapping);
}