//
// Created by Wayne on 16/1/5.
//

#ifndef REACHABILITY_SUMMARY_K_H
#define REACHABILITY_SUMMARY_K_H


#include <list>
#include <map>
#include <cassert>
#include <iostream>
#include <thread>
#include <sys/time.h>
#include <sys/resource.h>
#include "conditional_edge.h"
#include "library_graph.h"
#include "multilevel_queue.h"
#include "grammar.h"
#include "Logger.h"

using std::list;

class hocra_mapping;

pair<library_graph*, hocra_mapping*> analyze_libraries_k(int k, const library_graph *lib_gr, const vector<pair<int, int>> *conds, Logger &logger);
pair<library_graph*, hocra_mapping*> analyze_libraries_k_multi(int k, const library_graph *lib_gr, const vector<pair<int, int>> *conds);

class hocra_mapping {
    unordered_map<int, unordered_set<const conditional_edge*>> *record[7];
public:
    const unordered_map<int, unordered_set<const conditional_edge*>>& get_reocrd(int src, int sym) const {
        return record[sym][src];
    }
    hocra_mapping(int num_nodes) {
        for (int s = 0; s < 7; s++) {
            record[s] = new unordered_map<int, unordered_set<const conditional_edge*>>[num_nodes];
        }
    }
    void add1(int source, int target, int symbol, const conditional_edge *ce) {
        record[symbol][source][target].insert(ce);
    }
    void add2(int source, int target, int symbol, const conditional_edge *ce1, const conditional_edge *ce2) {
        auto &tmp = record[symbol][source][target];
        tmp.insert(ce1);
        tmp.insert(ce2);
    }
    void add3(int source, int target, int symbol, const conditional_edge *ce1, const conditional_edge *ce2, const conditional_edge *ce3) {
        auto &tmp = record[symbol][source][target];
        tmp.insert(ce1);
        tmp.insert(ce2);
        tmp.insert(ce3);
    }
    bool has(int src, int tgt, int sym) const {
        return (record[sym][src].count(tgt) != 0);
    }
    const unordered_set<const conditional_edge*>& get(int src, int tgt, int sym) const {
        return record[sym][src].at(tgt);
    }
};


#endif //REACHABILITY_SUMMARY_K_H
