//
// Created by Wayne on 16/1/23.
//

#include "summarize_tal.h"

using std::queue;
using std::cerr;
using std::endl;
using std::map;
using std::make_pair;
using std::unordered_set;

static unordered_set<int> special;

static inline void add_fst(int sym, queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    if (!gr->is_fst_redundant(src, tgt, sym)) {
        auto new_edge = new tal_fst_edge(sym, src, tgt);
        wl.push(new_edge);
        gr->add_fst_record(new_edge);
    }
}

static inline void add_snd(int sym, queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    if (!gr->is_fst_redundant(src, tgt, sym) && special.count(src) != 0 && special.count(tgt) != 0 && special.count(msrc) != 0 && special.count(mtgt) != 0 && !gr->is_snd_redundant(src, tgt, msrc, mtgt, sym)) {
        auto new_edge = new tal_snd_edge(sym, src, tgt, msrc, mtgt);
        wl.push(new_edge);
        gr->add_snd_record(new_edge);
    }
}

static inline void add_fst_S(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_S, wl, gr, src, tgt);
}

static inline void add_fst_L(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_L, wl, gr, src, tgt);
}

static inline void add_fst_R(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_R, wl, gr, src, tgt);
}

static inline void add_fst_D(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_D, wl, gr, src, tgt);
}

static inline void add_snd_S(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_S, wl, gr, src, tgt, msrc, mtgt);
}

static inline void add_snd_L(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_L, wl, gr, src, tgt, msrc, mtgt);
}

static inline void add_snd_R(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_R, wl, gr, src, tgt, msrc, mtgt);
}

tal_library_graph* analyze_libraries_tal(const tal_library_graph *lib_gr, Logger &logger) {
    auto start = clock();

    queue<const tal_edge_base*> work_list;
    tal_library_graph *tmp_gr = new tal_library_graph(lib_gr->get_num_nodes());

    vector<map<int, vector<pair<int, int>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    auto &lefts = lib_gr->get_left_parens();
    auto &rights = lib_gr->get_right_parens();
    for (auto it = lefts.begin(); it != lefts.end(); it++) {
        int left_paren = it->first;
        int right_paren = -it->first;
        if (rights.count(right_paren) != 0) {
            for (auto e1 : lefts.at(left_paren)) {
                for (auto e2 : rights.at(right_paren)) {
                    record[e1->get_target()][e2->get_source()].push_back(make_pair(e1->get_source(), e2->get_target()));
                }
            }
        }
    }

    special.clear();
    for (auto e : lib_gr->all_parens()) {
        if (is_left_paren(e->get_symbol())) {
            add_fst_L(work_list, tmp_gr, e->get_source(), e->get_target());
        } else if (is_right_paren(e->get_symbol())) {
            add_fst_R(work_list, tmp_gr, e->get_source(), e->get_target());
        } else {
            throw "bad implementation";
        }
        special.insert(e->get_source());
        special.insert(e->get_target());
    }
    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto fe : lib_gr->fst_edges_from(src, GRAM_S)) {
            add_fst_S(work_list, tmp_gr, src, fe->get_target());
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                add_fst_S(work_list, tmp_gr, p.first, p.second);
            }
        }
    }

    while (!work_list.empty()) {
        auto e = work_list.front();
        work_list.pop();

        int symbol = e->get_symbol();

        assert(!is_paren(symbol));

        if (e->get_order() == 1) {
            auto fe = (const tal_fst_edge*)e;
            int source = fe->get_source();
            int target = fe->get_target();
            tmp_gr->add_fst_edge(fe);

            if (symbol == GRAM_S) {
                // S -> @SS
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_S)) {
                    add_fst_S(work_list, tmp_gr, source, forward->get_target());
                }
                // S -> S@S
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_S)) {
                    add_fst_S(work_list, tmp_gr, backward->get_source(), target);
                }
                // L -> L@S
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_L)) {
                    add_fst_L(work_list, tmp_gr, backward->get_source(), target);
                }
                // L -> @SL
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_L)) {
                    add_fst_L(work_list, tmp_gr, source, forward->get_target());
                }
                // R -> R@S
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_R)) {
                    add_fst_R(work_list, tmp_gr, backward->get_source(), target);
                }
                // R -> @SR
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_R)) {
                    add_fst_R(work_list, tmp_gr, source, forward->get_target());
                }
                // S -> {i@S}i
                if (record[source].count(target) != 0) {
                    auto &pairs = record[source][target];
                    for (auto &p : pairs) {
                        add_fst_S(work_list, tmp_gr, p.first, p.second);
                    }
                }
            } else if (symbol == GRAM_L) {
                // L -> @LL
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_L)) {
                    add_fst_L(work_list, tmp_gr, source, forward->get_target());
                }
                // L -> L@L
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_L)) {
                    add_fst_L(work_list, tmp_gr, backward->get_source(), target);
                }
                // L -> @LS
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_S)) {
                    add_fst_L(work_list, tmp_gr, source, forward->get_target());
                }
                // L -> S@L
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_S)) {
                    add_fst_L(work_list, tmp_gr, backward->get_source(), target);
                }
                // D -> R@L
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_R)) {
                    add_fst_D(work_list, tmp_gr, backward->get_source(), target);
                }
            } else if (symbol == GRAM_R) {
                // R -> @RR
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_R)) {
                    add_fst_R(work_list, tmp_gr, source, forward->get_target());
                }
                // R -> R@R
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_R)) {
                    add_fst_R(work_list, tmp_gr, backward->get_source(), target);
                }
                // R -> @RS
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_S)) {
                    add_fst_R(work_list, tmp_gr, source, forward->get_target());
                }
                // R -> S@R
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_S)) {
                    add_fst_R(work_list, tmp_gr, backward->get_source(), target);
                }
                // D -> @RL
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_L)) {
                    add_fst_D(work_list, tmp_gr, source, forward->get_target());
                }
            } else if (symbol == GRAM_D) {

            } else {
                throw "bad implementation";
            }
        } else {
            throw "bad implementation";
        }
    }

    cerr << "FST edges: " << tmp_gr->get_num_edges() << endl;

    for (auto it = lefts.begin(); it != lefts.end(); it++) {
        int left_paren = it->first;
        int right_paren = -it->first;
        if (rights.count(right_paren) != 0) {
            for (auto e1 : lefts.at(left_paren)) {
                for (auto e2 : rights.at(right_paren)) {
                    add_snd_S(work_list, tmp_gr, e1->get_source(), e2->get_target(), e1->get_target(), e2->get_source());
                }
            }
        }
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto e = work_list.front();
        work_list.pop();

        int symbol = e->get_symbol();

        assert(!is_paren(symbol));

        if (e->get_order() == 2) {
            auto se = (const tal_snd_edge*)e;
            int source = se->get_source();
            int target = se->get_target();
            int mid_source = se->get_mid_source();
            int mid_target = se->get_mid_target();
            tmp_gr->add_snd_edge(se);

            if (symbol == GRAM_S) {
                // S -> UNPAIR(@SS)
                /*if (mid_source == mid_target) {
                    add_fst_S(work_list, tmp_gr, source, target);
                }*/
                // SS -> LL(@SS, S)
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, backward->get_source(), target, mid_source, mid_target);
                }
                // SS -> LR(@SS, S)
                for (auto forward : tmp_gr->fst_edges_from(mid_source, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, source, target, forward->get_target(), mid_target);
                }
                // SS -> RL(@SS, S)
                for (auto backward : tmp_gr->fst_edges_to(mid_target, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, source, target, mid_source, backward->get_source());
                }
                // SS -> RR(@SS, S)
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, source, forward->get_target(), mid_source, mid_target);
                }
                // LL -> LL(@SS, L)
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_L)) {
                    add_snd_L(work_list, tmp_gr, backward->get_source(), target, mid_source, mid_target);
                }
                // LL -> RR(@SS, L)
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_L)) {
                    add_snd_L(work_list, tmp_gr, source, forward->get_target(), mid_source, mid_target);
                }
                // RR -> LL(@SS, R)
                for (auto backward : tmp_gr->fst_edges_to(source, GRAM_R)) {
                    add_snd_R(work_list, tmp_gr, backward->get_source(), target, mid_source, mid_target);
                }
                // RR -> RR(@SS, R)
                for (auto forward : tmp_gr->fst_edges_from(target, GRAM_R)) {
                    add_snd_R(work_list, tmp_gr, source, forward->get_target(), mid_source, mid_target);
                }
                // SS -> ADJ(@SS, SS)
                for (auto inner : tmp_gr->snd_edges_of(mid_source, mid_target, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, source, target, inner->get_mid_source(), inner->get_mid_target());
                }
                // SS -> ADJ(SS, @SS)
                for (auto outer : tmp_gr->snd_edges_of_mid(source, target, GRAM_S)) {
                    add_snd_S(work_list, tmp_gr, outer->get_source(), outer->get_target(), mid_source, mid_target);
                }
            } else if (symbol == GRAM_L) {
                // L -> UNPAIR(@LL)
                /*if (mid_source == mid_target) {
                    add_fst_L(work_list, tmp_gr, source, target);
                }*/
                // LL -> LL(@LL, L)
                /*for (auto backward : tmp_gr->fst_edges_to(source, GRAM_L)) {
                    add_snd_L(work_list, tmp_gr, backward->get_source(), target, mid_source, mid_target);
                }*/
                // LL -> RR(@LL, L)
                /*for (auto forward : tmp_gr->fst_edges_from(target, GRAM_L)) {
                    add_snd_L(work_list, tmp_gr, source, forward->get_target(), mid_source, mid_target);
                }*/
            } else if (symbol == GRAM_R) {
                // R -> UNPAIR(@RR)
                /*if (mid_source == mid_target) {
                    add_fst_R(work_list, tmp_gr, source, target);
                }*/
                // RR -> LL(@RR, R)
                /*for (auto backward : tmp_gr->fst_edges_to(source, GRAM_R)) {
                    add_snd_R(work_list, tmp_gr, backward->get_source(), target, mid_source, mid_target);
                }*/
                // RR -> RR(@RR, R)
                /*for (auto forward : tmp_gr->fst_edges_from(target, GRAM_R)) {
                    add_snd_R(work_list, tmp_gr, source, forward->get_target(), mid_source, mid_target);
                }*/
            } else {
                throw "bad implementation";
            }
        } else {
            throw "bad implementation";
        }
    }

    int lib_nodes = lib_gr->get_num_nodes();
    vector<bool> vis(lib_nodes, false);
    queue<int> q;
    for (int n = 0; n < lib_nodes; n++) {
        if (!vis[n] && lib_gr->is_return_node(n)) {
            vis[n] = true;
            q.push(n);
            while (!q.empty()) {
                int u = q.front();
                q.pop();
                for (auto fe : lib_gr->fst_edges_from(u, GRAM_S)) {
                    int v = fe->get_target();
                    if (!vis[v]) {
                        vis[v] = true;
                        q.push(v);
                    }
                }
            }
        }
    }
    for (int n = 0; n < lib_nodes; n++) {
        if (vis[n] && lib_gr->is_call_node(n)) {
            tmp_gr->add_chain_node(n);
        }
    }
    for (auto &p : lib_gr->get_lib_info()) {
        int start = p.first;
        int len = p.second;
        bool chk = true;
        for (int n = start; n < start + len && chk; n++) {
            if (tmp_gr->is_chain_node(n)) {
                chk = false;
            }
        }
        if (!chk) {
            for (int n = start; n < start + len; n++) {
                if (lib_gr->is_entry_node(n) || lib_gr->is_exit_node(n)) {
                    tmp_gr->add_chain_node(n);
                }
            }
        }
    }

    cerr << "Chain nodes: " << tmp_gr->get_chain_nodes().size() << endl;
    cerr << "Total edges: " << tmp_gr->get_num_edges() << endl;

    auto end = clock();
    cerr << "Summarize time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    logger.log("#Chain Node in the Summary: ");
    logger.log(tmp_gr->get_chain_nodes().size());
    logger.log("\n");
    logger.log("#Edge in the Summary: ");
    logger.log(tmp_gr->get_num_edges());
    logger.log("\n");
    logger.log("Time of Summarization: ");
    logger.log((end - start) * 1.0 / CLOCKS_PER_SEC);
    logger.log("\n");

    return tmp_gr;
}