//
// Created by Wayne on 16/1/23.
//

#ifndef REACHABILITY_TAL_LIBRARY_GRAPH_H
#define REACHABILITY_TAL_LIBRARY_GRAPH_H


#include <vector>
#include <unordered_map>
#include <unordered_set>
#include <fstream>
#include <iostream>
#include <sstream>
#include "tal_edge.h"
#include "grammar.h"

using std::vector;
using std::unordered_map;
using std::unordered_set;
using std::string;
using std::pair;

class tal_library_graph {
    int num_nodes;
    int num_edges;
    vector<const tal_fst_edge*> parens;
    vector<const tal_fst_edge*> *fst_forward[7], *fst_backward[7];
    unordered_set<int> *fst_record[7];
    vector<const tal_snd_edge*> *snd_forward[3], *snd_backward[3], *snd_forward_mid[3], *snd_backward_mid[3];
    unordered_map<int, unordered_map<int, unordered_set<int>>> *snd_record[3];
    unordered_map<int, vector<const tal_snd_edge*>> *snd_of[3], *snd_of_mid[3];
    unordered_map<int, vector<const tal_fst_edge*>> left_parens, right_parens;
    unordered_set<int> call_nodes, return_nodes, entry_nodes, exit_nodes;
    unordered_set<int> chain_nodes;
    vector<pair<int, int>> lib_info;
public:
    tal_library_graph(int num_nodes);
    int get_num_nodes() const;
    int get_num_edges() const;
    const vector<const tal_fst_edge*>& fst_edges_from(int source, int symbol) const;
    const vector<const tal_fst_edge*>& fst_edges_to(int target, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_from(int source, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_to(int target, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_from_mid(int source, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_to_mid(int target, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_of(int source, int target, int symbol) const;
    const vector<const tal_snd_edge*>& snd_edges_of_mid(int source, int target, int symbol) const;
    const vector<const tal_fst_edge*>& all_left_parens_of(int symbol) const;
    const vector<const tal_fst_edge*>& all_right_parens_of(int symbol) const;
    const vector<const tal_fst_edge*>& all_parens() const;
    void add_fst_edge(const tal_fst_edge *fe);
    void add_fst_record(const tal_fst_edge *fe);
    void add_snd_edge(const tal_snd_edge *se);
    void add_snd_record(const tal_snd_edge *se);
    void add_paren(const tal_fst_edge *e);
    bool is_fst_redundant(int source, int target, int symbol) const;
    bool is_snd_redundant(int source, int target, int mid_source, int mid_target, int symbol) const;
    const unordered_map<int, vector<const tal_fst_edge*>>& get_left_parens() const;
    const unordered_map<int, vector<const tal_fst_edge*>>& get_right_parens() const;
    bool is_call_node(int n) const;
    bool is_return_node(int n) const;
    bool is_entry_node(int n) const;
    bool is_exit_node(int n) const;
    void add_chain_node(int n);
    bool is_chain_node(int n) const;
    void load_lib_info(const string &file_name);
    const vector<pair<int, int>>& get_lib_info() const;
    const unordered_set<int>& get_chain_nodes() const;
public:
    static vector<const tal_fst_edge*> FST_EMPTY;
    static vector<const tal_snd_edge*> SND_EMPTY;
public:
    static tal_library_graph* load2(const string &file_name);
};


#endif //REACHABILITY_TAL_LIBRARY_GRAPH_H
